""" Unit tests for the fileIO module """

import os
import numpy as np

from tables_io import io

h5_data_file = 'tests/data/pandas_test_hdf5.h5'
parquet_data_file = 'tests/data/parquet_test.parquet'
no_group_file = 'tests/data/no_groupname_test.hdf5'

test_outfile = './test_out.h5'


def test_pandas_readin():
    """ Test the pandas reading """
    _ = io.readH5ToDict(h5_data_file)
    _ = io.readPqToDict(parquet_data_file)


def test_no_groupname():
    """ Test the load_training_data function for a file with no groupname """
    _ = io.readHdf5ToDict(no_group_file, groupname=None)

def test_get_input_data_length_hdf5():
    """ Test the get_input_data_size_hdf5 function """
    assert io.getInputDataLengthHdf5(no_group_file) == 10
    try:
        _ = io.getInputDataLengthHdf5(h5_data_file, 'df')
    except ValueError:
        pass
    else:
        raise ValueError("Failed to catch ValueError for mismatched column lengths")


def test_iter_chunk_hdf5_data():
    """ Test the hdf5 data chunk iterator """
    for itr in io.iterHdf5ToDict(no_group_file, chunk_size=2):
        for val in itr[2].values():
            assert np.size(val) == 2

    for itr in io.iterHdf5ToDict(no_group_file, chunk_size=3):
        for val in itr[2].values():
            assert np.size(val) <= 3


def test_write_output_file():
    """ Test writing an output file """
    npdf = 40
    nbins = 21
    pz_pdf = np.random.uniform(size=(npdf, nbins))
    zgrid = np.linspace(0, 4, nbins)
    zmode = zgrid[np.argmax(pz_pdf, axis=1)]

    data_dict = dict(zmode=zmode, pz_pdf=pz_pdf)

    group, outf = io.initializeHdf5Write(test_outfile, 'data', photoz_mode=((npdf,), 'f4'), photoz_pdf=((npdf, nbins), 'f4'))
    io.writeDictToHdf5Chunk(group, data_dict, 0, npdf, zmode='photoz_mode', pz_pdf='photoz_pdf')
    io.finalizeHdf5Write(outf, 'md', zgrid=zgrid)

    os.unlink(test_outfile)



"""IO Functions for tables_io"""

import os
from collections import OrderedDict

import numpy as np

from .lazy_modules import pd, pq, h5py, apTable, fits
from .lazy_modules import HAS_TABLES, HAS_HDF5

from .arrayUtils import getGroupInputDataLength

from .types import ASTROPY_FITS, ASTROPY_HDF5, NUMPY_HDF5, NUMPY_FITS, PANDAS_HDF5, PANDAS_PARQUET,\
     NATIVE_FORMAT, FILE_FORMAT_SUFFIXS, FILE_FORMAT_SUFFIX_MAP, DEFAULT_TABLE_KEY,\
     NATIVE_TABLE_TYPE, AP_TABLE, PD_DATAFRAME,\
     fileType, tableType, istablelike, istabledictlike

from .convUtils import dataFrameToDict, hdf5GroupToDict, convert



### I. Iteration functions


### I A. HDF5 partial read/write functions

def readHdf5DatasetToArray(dataset, start=None, end=None):
    """
    Reads part of a hdf5 dataset into a `numpy.array`

    Parameters
    ----------
    dataset : `h5py.Dataset`
        The input dataset

    start : `int` or `None`
        Starting row

    end : `int` or `None`
        Ending row

    Returns
    -------
    out : `numpy.array` or `list` of `numpy.array`
        Something that pandas can handle
    """
    if start is None or end is None:
        return np.array(dataset)
    return np.array(dataset[start:end])


def getInputDataLengthHdf5(filepath, groupname=None):
    """ Open an HDF5 file and return the size of a group

    Parameters
    ----------
    filepath: `str`
        Path to input file

    groupname : `str` or `None`
        The groupname for the data


    Returns
    -------
    length : `int`
        The length of the data

    Notes
    -----
    For a multi-D array this return the length of the first axis
    and not the total size of the array.

    Normally that is what you want to be iterating over.
    """
    hg, infp = readHdf5Group(filepath, groupname)
    nrow = getGroupInputDataLength(hg)
    infp.close()
    return nrow


def initializeHdf5Write(filepath, groupname=None, **kwds):
    """ Prepares an hdf5 file for output

    Parameters
    ----------
    filepath : `str`
        The output file name
    groupname : `str` or `None`
        The output group name

    Returns
    -------
    group : `h5py.File` or `h5py.Group`
        The group to write to
    fout : `h5py.File`
        The output file

    Notes
    -----
    The keywords should be used to create_datasets within the hdf5 file.
    Each keyword should provide a tuple of ( (shape), (dtype) )

    shape : `tuple` ( `int` )
        The shape of the data for this dataset
    dtype : `str`
        The data type for this dataset

    For exmaple
    `initialize_writeout('test.hdf5', scalar=((100000,), 'f4'), vect=((100000, 3), 'f4'))`

    Would initialize an hdf5 file with two datasets, with shapes and data types as given
    """
    outdir = os.path.dirname(os.path.abspath(filepath))
    if not os.path.exists(outdir):  #pragma: no cover
        os.makedirs(outdir, exist_ok=True)
    outf = h5py.File(filepath, "w")
    if groupname is None:  #pragma: no cover
        group = outf
    else:
        group = outf.create_group(groupname)

    for k, v in kwds.items():
        group.create_dataset(k, v[0], v[1])
    return group, outf


def writeDictToHdf5Chunk(fout, odict, start, end, **kwds):
    """ Writes a data chunk to an hdf5 file

    Parameters
    ----------
    fout : `h5py.File`
        The file

    odict : `OrderedDict`, (`str`, `numpy.array`)
        The data being written

    start : `int`
        Starting row number to place the data

    end : `int`
        Ending row number to place the data

    Notes
    -----
    The kwds can be used to control the output locations, i.e., to
    rename the columns in data_dict when they good into the output file.

    For each item in data_dict, the output location is set as

    `k_out = kwds.get(key, key)`

    This will check the kwds to see if they contain `key` and if so, will
    return the corresponding value.  Otherwise it will just return `key`.

    I.e., if `key` is present in kwds in will override the name.
    """
    for key, val in odict.items():
        k_out = kwds.get(key, key)
        fout[k_out][start:end] = val


def finalizeHdf5Write(fout, groupname=None, **kwds):
    """ Write any last data and closes an hdf5 file

    Parameters
    ----------
    fout : `h5py.File`
        The file

    Notes
    -----
    The keywords can be used to write additional data
    """
    if groupname is None:  #pragma: no cover
        group = fout
    else:
        group = fout.create_group(groupname)
    for k, v in kwds.items():
        group[k] = v
    fout.close()


def iterHdf5ToDict(filepath, chunk_size=100_000, groupname=None):
    """
    iterator for sending chunks of data in hdf5.

    Parameters
    ----------
      filepath: input file name (str)
      chunk_size: size of chunk to iterate over (int)

    Returns
    -------
    output:
        iterator chunk

    Currently only implemented for hdf5, returns `tuple`
        start: start index (int)
        end: ending index (int)
        data: dictionary of all data from start:end (dict)
    """
    f, infp = readHdf5Group(filepath, groupname)
    num_rows = getGroupInputDataLength(f)
    data = OrderedDict()
    for i in range(0, num_rows, chunk_size):
        start = i
        end = min(i+chunk_size, num_rows)
        for key, val in f.items():
            data[key] = readHdf5DatasetToArray(val, start, end)
        yield start, end, data
    infp.close()


def iterH5ToDataFrame(filepath, chunk_size=100_000, groupname=None):
    """
    iterator for sending chunks of data in hdf5.

    Parameters
    ----------
      filepath: input file name (str)
      chunk_size: size of chunk to iterate over (int)

    Returns
    -------
    output:
        iterator chunk

    Currently only implemented for hdf5, returns `tuple`
        start: start index (int)
        end: ending index (int)
        data: pandas.DataFrame of all data from start:end (dict)
    """
    raise NotImplementedError("iterH5ToDataFrame")

    # This does't work b/c of the difference in structure
    #f, infp = readHdf5Group(filepath, groupname)
    #num_rows = getGroupInputDataLength(f)
    #for i in range(0, num_rows, chunk_size):
    #    start = i
    #    end = i+chunk_size
    #    if end > num_rows:
    #        end = num_rows
    #    data = pd.read_hdf(filepath, start=start, stop=end)
    #    yield start, end, data
    #infp.close()


def iterPqToDataFrame(filepath):
    """
    iterator for sending chunks of data in parquet

    Parameters
    ----------
      filepath: input file name (str)

    Returns
    -------
    output:
        iterator chunk

    Currently only implemented for hdf5, returns `tuple`
        start: start index (int)
        end: ending index (int)
        data: `pandas.DataFrame` of all data from start:end (dict)
    """
    raise NotImplementedError("iterPqToDataFrame")


### II.   Reading and Writing Files

### II A.  Reading and Writing `astropy.table.Table` to/from FITS files

def writeApTablesToFits(tables, filepath, **kwargs):
    """
    Writes a dictionary of `astropy.table.Table` to a single FITS file

    Parameters
    ----------
    tables : `dict` of `astropy.table.Table`
        Keys will be HDU names, values will be tables

    filepath: `str`
        Path to output file

    kwargs are passed to `astropy.io.fits.writeto` call.
    """
    out_list = [fits.PrimaryHDU()]
    for k, v in tables.items():
        hdu = fits.table_to_hdu(v)
        hdu.name = k
        out_list.append(hdu)
    hdu_list = fits.HDUList(out_list)
    hdu_list.writeto(filepath, **kwargs)


def readFitsToApTables(filepath):
    """
    Reads `astropy.table.Table` objects from a FITS file.

    Parameters
    ----------
    filepath: `str`
        Path to input file

    Returns
    -------
    tables : `OrderedDict` of `astropy.table.Table`
        Keys will be HDU names, values will be tables
    """
    fin = fits.open(filepath)
    tables = OrderedDict()
    for hdu in fin[1:]:
        tables[hdu.name.lower()] = apTable.Table.read(filepath, hdu=hdu.name)
    return tables


### II A'.  Reading and Writing `` to/from FITS files

def writeRecarraysToFits(recarrays, filepath, **kwargs):
    """
    Writes a dictionary of `np.recarray` to a single FITS file

    Parameters
    ----------
    recarrays  : `dict` of `np.recarray`
        Keys will be HDU names, values will be tables

    filepath: `str`
        Path to output file

    kwargs are passed to `astropy.io.fits.writeto` call.
    """
    out_list = [fits.PrimaryHDU()]
    for k, v in recarrays.items():
        hdu = fits.BinTableHDU.from_columns(v.columns)
        hdu.name = k
        out_list.append(hdu)
    hdu_list = fits.HDUList(out_list)
    hdu_list.writeto(filepath, **kwargs)


def readFitsToRecarrays(filepath):
    """
    Reads `np.recarray` objects from a FITS file.

    Parameters
    ----------
    filepath: `str`
        Path to input file

    Returns
    -------
    tables : `OrderedDict` of `np.recarray`
        Keys will be HDU names, values will be tables
    """
    fin = fits.open(filepath)
    tables = OrderedDict()
    for hdu in fin[1:]:
        tables[hdu.name.lower()] = hdu.data
    return tables

### II B.  Reading and Writing `astropy.table.Table` to/from `hdf5`

def writeApTablesToHdf5(tables, filepath, **kwargs):
    """
    Writes a dictionary of `astropy.table.Table` to a single hdf5 file

    Parameters
    ----------
    tables : `dict` of `astropy.table.Table`
        Keys will be passed to 'path' parameter

    filepath: `str`
        Path to output file

    kwargs are passed to `astropy.table.Table` call.
    """
    if not HAS_HDF5:  #pragma: no cover
        raise ImportError("h5py is not available, can't save to hdf5")

    for k, v in tables.items():
        v.write(filepath, path=k, append=True, format='hdf5', **kwargs)


def readHdf5ToApTables(filepath):
    """
    Reads `astropy.table.Table` objects from an hdf5 file.

    Parameters
    ----------
    filepath: `str`
        Path to input file

     Returns
    -------
    tables : `OrderedDict` of `astropy.table.Table`
        Keys will be 'paths', values will be tables
    """
    fin = h5py.File(filepath)
    tables = OrderedDict()
    for k in fin.keys():
        tables[k] = apTable.Table.read(filepath, path=k, format='hdf5')
    return tables


### II C.  Reading and Writing `OrderedDict`, (`str`, `numpy.array`) to/from `hdf5`

def readHdf5Group(filepath, groupname=None):
    """ Read and return group from an hdf5 file.

    Parameters
    ----------
    filepath : `str`
        File in question
    groupname : `str` or `None`
        For hdf5 files, the groupname for the data

    Returns
    -------
    grp : `h5py.Group` or `h5py.File`
        The requested group
    infp : `h5py.File`
        The input file (returned so that the used can explicitly close the file)
    """
    infp = h5py.File(filepath, "r")
    if groupname is None:  #pragma: no cover
        return infp, infp
    return infp[groupname], infp


def readHdf5GroupToDict(hg, start=None, end=None):
    """
    Reads `numpy.array` objects from an hdf5 file.

    Parameters
    ----------
    filepath: `str`
        Path to input file

    Returns
    -------
    tables : `OrderedDict` of `numpy.array`
        Keys will be 'paths', values will be tables
    """
    # pylint: disable=unused-argument
    if isinstance(hg, h5py.Dataset):
        return readHdf5DatasetToArray(hg, start, end)
    return OrderedDict([(key, readHdf5DatasetToArray(val, start, end)) for key, val in hg.items()])


def writeDictToHdf5(odict, filepath, groupname, **kwargs):
    """
    Writes a dictionary of `numpy.array` to a single hdf5 file

    Parameters
    ----------
    odict : `Mapping`, (`str`, `numpy.array`)
        The data being written

    filepath: `str`
        Path to output file

    groupname : `str` or `None`
        The groupname for the data
    """
    # pylint: disable=unused-argument
    fout = h5py.File(filepath, 'a')
    if groupname is None:  #pragma: no cover
        group = fout
    else:
        group = fout.create_group(groupname)
    for key, val in odict.items():
        try:
            group.create_dataset(key, dtype=val.dtype, data=val.data)
        except Exception as msg:  #pragma: no cover
            print(f"Warning.  Failed to convert column {str(msg)}")
    fout.close()


def writeDictsToHdf5(odicts, filepath):
    """
    Writes a dictionary of `numpy.array` to a single hdf5 file

    Parameters
    ----------
    odicts : `OrderedDict`, (`str`, `Tablelike`)
        The data being written

    filepath: `str`
        Path to output file
    """
    try:
        os.unlink(filepath)
    except FileNotFoundError:
        pass
    for key, val in odicts.items():
        writeDictToHdf5(val, filepath, key)


def readHdf5ToDicts(filepath):
    """
    Reads `numpy.array` objects from an hdf5 file.

    Parameters
    ----------
    filepath: `str`
        Path to input file

    Returns
    -------
    dicts : `OrderedDict`, (`str`, `OrderedDict`, (`str`, `numpy.array`) )
        The data
    """
    fin = h5py.File(filepath)
    return OrderedDict([(key, readHdf5GroupToDict(val)) for key, val in fin.items()])



### II C.  Reading and Writing `pandas.DataFrame` to/from `hdf5`

def readHdf5ToDataFrame(filepath, key=None):
    """
    Reads `pandas.DataFrame` objects from an hdf5 file.

    Parameters
    ----------
    filepath: `str`
        Path to input file
    key : `str` or `None`
        The key in the hdf5 file

    Returns
    -------
    df : `pandas.DataFrame`
        The dataframe
    """
    if not HAS_HDF5:  #pragma: no cover
        raise ImportError("h5py is not available, can't read hdf5")
    if not HAS_TABLES:  #pragma: no cover
        raise ImportError("tables is not available, can't read hdf5 to dataframe")

    return pd.read_hdf(filepath, key)


def readH5ToDataFrames(filepath):
    """ Open an h5 file and and return a dictionary of `pandas.DataFrame`

    Parameters
    ----------
    filepath: `str`
        Path to input file

    Returns
    -------
    tab : `OrderedDict` (`str` : `pandas.DataFrame`)
       The data

    Notes
    -----
    We are using the file suffix 'h5' to specify 'hdf5' files written from DataFrames using `pandas`
    They have a different structure than 'hdf5' files written with `h5py` or `astropy.table`
    """
    if not HAS_TABLES:  #pragma: no cover
        raise ImportError("tables is not available, can't read hdf5 to dataframe")
    fin = h5py.File(filepath)
    return OrderedDict([(key, readHdf5ToDataFrame(filepath, key=key)) for key in fin.keys()])


def writeDataFramesToH5(dataFrames, filepath):
    """
    Writes a dictionary of `pandas.DataFrame` to a single hdf5 file

    Parameters
    ----------
    dataFrames : `dict` of `pandas.DataFrame`
        Keys will be passed to 'key' parameter

    filepath: `str`
        Path to output file
    """
    if not HAS_HDF5:  #pragma: no cover
        raise ImportError("h5py is not available, can't write hdf5")
    if not HAS_TABLES:  #pragma: no cover
        raise ImportError("tables is not available, can't write hdf5")

    for key, val in dataFrames.items():
        val.to_hdf(filepath, key)




### II E.  Reading and Writing `pandas.DataFrame` to/from `parquet`

def readPqToDataFrame(filepath):
    """
    Reads a `pandas.DataFrame` object from an parquet file.

    Parameters
    ----------
    filepath: `str`
        Path to input file

    Returns
    -------
    df : `pandas.DataFrame`
        The data frame
    """
    return pd.read_parquet(filepath, engine='pyarrow')


def writeDataFramesToPq(dataFrames, filepath, **kwargs):
    """
    Writes a dictionary of `pandas.DataFrame` to a parquet files

    Parameters
    ----------
    tables : `dict` of `pandas.DataFrame`
        Keys will be passed to 'path' parameter

    filepath: `str`
        Path to output file

    """
    for k, v in dataFrames.items():
        _ = v.to_parquet(f"{filepath}{k}.pq", **kwargs)


def readPqToDataFrames(basepath, keys=None, allow_missing_keys=False):
    """
    Reads `pandas.DataFrame` objects from an parquet file.

    Parameters
    ----------
    basepath: `str`
        Path to input file

    keys : `list`
        Keys for the input objects.  Used to complete filepaths

    allow_missing_keys: `bool`
        If False will raise FileNotFoundError if a key is missing

    Returns
    -------
    tables : `OrderedDict` of `pandas.DataFrame`
        Keys will be taken from keys
    """
    if keys is None:  #pragma: no cover
        keys = [""]
    dataframes = OrderedDict()
    for key in keys:
        try:
            dataframes[key] = readPqToDataFrame(f"{basepath}{key}.pq")
        except FileNotFoundError as msg:  #pragma: no cover
            if allow_missing_keys:
                continue
            raise FileNotFoundError from msg
    return dataframes


### II F.  Reading and Writing to `OrderedDict`, (`str`, `numpy.array`)

def readPqToDict(filepath, columns=None):
    """ Open a parquet file and return a dictionary of `numpy.array`

    Parameters
    ----------
    filepath: `str`
        Path to input file

    columns : `list` (`str`) or `None`
        Names of the columns to read, `None` will read all the columns

    Returns
    -------
    tab : `OrderedDict` (`str` : `numpy.array`)
       The data
    """
    tab = pq.read_table(filepath, columns=columns)
    return OrderedDict([(c_name, col.to_numpy()) for c_name, col in zip(tab.column_names, tab.itercolumns())])


def readH5ToDict(filepath, groupname=None):
    """ Open an h5 file and and return a dictionary of `numpy.array`

    Parameters
    ----------
    filepath: `str`
        Path to input file

    groupname : `str` or `None`
        The group with the data

    Returns
    -------
    tab : `OrderedDict` (`str` : `numpy.array`)
       The data

    Notes
    -----
    We are using the file suffix 'h5' to specify 'hdf5' files written from DataFrames using `pandas`
    They have a different structure than 'hdf5' files written with `h5py` or `astropy.table`
    """
    df = readHdf5ToDataFrame(filepath, groupname)
    return dataFrameToDict(df)


def readHdf5ToDict(filepath, groupname=None):
    """ Read in h5py hdf5 data, return a dictionary of all of the keys

    Parameters
    ----------
    filepath: `str`
        Path to input file

    groupname : `str` or `None`
        The groupname for the data

    Returns
    -------
    tab : `OrderedDict` (`str` : `numpy.array`)
       The data

    Notes
    -----
    We are using the file suffix 'hdf5' to specify 'hdf5' files written with `h5py` or `astropy.table`
    They have a different structure than 'h5' files written `panda`
    """
    hg, infp = readHdf5Group(filepath, groupname)
    data = hdf5GroupToDict(hg)
    infp.close()
    return data


### II G.  Top-level interface functions

def io_open(filepath, fmt=None, **kwargs):
    """ Open a file

    Parameters
    ----------
    filepath : `str`
        File to load
    fmt : `str` or `None`
        File format, if `None` it will be taken from the file extension

    Returns
    -------
    file
    """
    fType = fileType(filepath, fmt)
    if fType in [ASTROPY_FITS, NUMPY_FITS]:
        return fits.open(filepath, **kwargs)
    if fType in [ASTROPY_HDF5, NUMPY_HDF5, PANDAS_HDF5]:
        return h5py.File(filepath, **kwargs)
    if fType in [PANDAS_PARQUET]:
        #basepath = os.path.splitext(filepath)[0]
        return pq.ParquetFile(filepath, **kwargs)
    raise TypeError(f"Unsupported FileType {fType}")  #pragma: no cover


def readNative(filepath, fmt=None, keys=None, allow_missing_keys=False):
    """ Read a file to the corresponding table type

    Parameters
    ----------
    filepath : `str`
        File to load
    fmt : `str` or `None`
        File format, if `None` it will be taken from the file extension
    keys : `list` or `None`
        For parquet files we must specify with keys to read, as each is in its own file
    allow_missing_keys : `bool`
        If False will raise FileNotFoundError if a key is missing

    Returns
    -------
    data : `OrderedDict` ( `str` -> `Tablelike` )
        The data

    """
    fType = fileType(filepath, fmt)
    if fType == ASTROPY_FITS:
        return readFitsToApTables(filepath)
    if fType == ASTROPY_HDF5:
        return readHdf5ToApTables(filepath)
    if fType == NUMPY_HDF5:
        return readHdf5ToDicts(filepath)
    if fType == NUMPY_FITS:
        return readFitsToRecarrays(filepath)
    if fType == PANDAS_HDF5:
        return readH5ToDataFrames(filepath)
    if fType == PANDAS_PARQUET:
        basepath = os.path.splitext(filepath)[0]
        return readPqToDataFrames(basepath, keys, allow_missing_keys)
    raise TypeError(f"Unsupported FileType {fType}")  #pragma: no cover


def read(filepath, tType=None, fmt=None, keys=None, allow_missing_keys=False):
    """ Read a file to the corresponding table type

    Parameters
    ----------
    filepath : `str`
        File to load
    tType : `int` or `None`
        Table type, if `None` this will use `readNative`
    fmt : `str` or `None`
        File format, if `None` it will be taken from the file extension
    keys : `list` or `None`
        For parquet files we must specify with keys to read, as each is in its own file
    allow_missing_keys : `bool`
        If False will raise FileNotFoundError if a key is missing

    Returns
    -------
    data : `OrderedDict` ( `str` -> `Tablelike` )
        The data

    """
    odict = readNative(filepath, fmt, keys, allow_missing_keys)
    if len(odict) == 1:
        for defName in ['', None, '__astropy_table__', 'data']:
            if defName in odict:
                odict = odict[defName]
    if tType is None:  #pragma: no cover
        return odict
    return convert(odict, tType)


def iteratorNative(filepath, fmt=None, **kwargs):
    """ Read a file to the corresponding table type and iterate over the file

    Parameters
    ----------
    filepath : `str`
        File to load
    fmt : `str` or `None`
        File format, if `None` it will be taken from the file extension

    Returns
    -------
    data : `TableLike`
        The data

    Notes
    -----
    The kwargs are used passed to the specific iterator type

    """
    fType = fileType(filepath, fmt)
    funcDict = {NUMPY_HDF5:iterHdf5ToDict,
                PANDAS_HDF5:iterH5ToDataFrame,
                PANDAS_PARQUET:iterPqToDataFrame}

    try:
        theFunc = funcDict[fType]
        return theFunc(filepath, **kwargs)
    except KeyError as msg:
        raise NotImplementedError(f"Unsupported FileType for iterateNative {fType}") from msg #pragma: no cover


def iterator(filepath, tType=None, fmt=None, **kwargs):
    """ Read a file to the corresponding table type iterate over the file

    Parameters
    ----------
    filepath : `str`
        File to load
    tType : `int` or `None`
        Table type, if `None` this will use `readNative`
    fmt : `str` or `None`
        File format, if `None` it will be taken from the file extension
    groupname : `str` or `None`
        For hdf5 files, the groupname for the data

    Returns
    -------
    data : `OrderedDict` ( `str` -> `Tablelike` )
        The data

    """
    for start, stop, data in iteratorNative(filepath, fmt, **kwargs):
        yield start, stop, convert(data, tType)


def writeNative(odict, basename):
    """ Write a file or files with tables

    Parameters
    ----------
    odict : `OrderedDict`, (`str`, `Tablelike`)
        The data to write
    basename : `str`
        Basename for the file to write.  The suffix will be applied based on the object type.
    """
    istable = False
    if istablelike(odict):
        istable = True
        tType = tableType(odict)
    elif istabledictlike(odict):
        tType = tableType(list(odict.values())[0])
    elif not odict:  #pragma: no cover
        return None
    else:  #pragma: no cover
        raise TypeError(f"Can not write object of type {type(odict)}")

    try:
        fType = NATIVE_FORMAT[tType]
    except KeyError as msg:  #pragma: no cover
        raise KeyError(f"No native format for table type {tType}") from msg
    fmt = FILE_FORMAT_SUFFIX_MAP[fType]
    filepath = basename + '.' + fmt

    if istable:
        odict = OrderedDict([(DEFAULT_TABLE_KEY[fmt], odict)])

    try:
        os.unlink(filepath)
    except FileNotFoundError:
        pass

    if fType == ASTROPY_HDF5:
        writeApTablesToHdf5(odict, filepath)
        return filepath
    if fType == NUMPY_HDF5:
        writeDictsToHdf5(odict, filepath)
        return filepath
    if fType == NUMPY_FITS:
        writeRecarraysToFits(odict, filepath)
        return filepath
    if fType == PANDAS_PARQUET:
        writeDataFramesToPq(odict, basename)
        return basename
    raise TypeError(f"Unsupported Native file type {fType}")  #pragma: no cover


def write(obj, basename, fmt=None):
    """ Write a file or files with tables

    Parameters
    ----------
    obj : `Tablelike` or `TableDictLike`
        The data to write
    basename : `str`
        Basename for the file to write.  The suffix will be applied based on the object type.
    fmt : `str` or `None`
        The output file format, If `None` this will use `writeNative`
    """
    if fmt is None:
        splitpath = os.path.splitext(basename)
        if not splitpath[1]:
            return writeNative(obj, basename)
        basename = splitpath[0]
        fmt = splitpath[1][1:]

    try:
        fType = FILE_FORMAT_SUFFIXS[fmt]
    except KeyError as msg:  #pragma: no cover
        raise KeyError(f"Unknown file format {fmt}, options are {list(FILE_FORMAT_SUFFIXS.keys())}") from msg

    if istablelike(obj):
        odict = OrderedDict([(DEFAULT_TABLE_KEY[fmt], obj)])
    elif istabledictlike(obj):
        odict = obj
    elif not obj:
        return None
    else:
        raise TypeError(f"Can not write object of type {type(obj)}")

    if fType in [ASTROPY_HDF5, NUMPY_HDF5, NUMPY_FITS, PANDAS_PARQUET]:
        try:
            nativeTType = NATIVE_TABLE_TYPE[fType]
        except KeyError as msg:  #pragma: no cover
            raise KeyError(f"Native file type not known for {fmt}") from msg

        forcedOdict = convert(odict, nativeTType)
        return writeNative(forcedOdict, basename)

    if fType == ASTROPY_FITS:
        forcedOdict = convert(odict, AP_TABLE)
        filepath = f"{basename}.fits"
        writeApTablesToFits(forcedOdict, filepath)
        return filepath
    if fType == PANDAS_HDF5:
        forcedOdict = convert(odict, PD_DATAFRAME)
        filepath = f"{basename}.h5"
        writeDataFramesToH5(forcedOdict, filepath)
        return filepath

    raise TypeError(f"Unsupported File type {fType}")  #pragma: no cover



""" Unit tests for the fileIO module """

import numpy as np

from tables_io import arrayUtils, types
from tables_io.lazy_modules import lazyImport


def test_array_length():
    """ Test the pandas reading """
    assert arrayUtils.arrayLength(4) == 0
    assert arrayUtils.arrayLength(np.ones(5)) == 5
    assert arrayUtils.arrayLength(np.ones((5, 5, 5))) == 5
    assert arrayUtils.arrayLength([3, 4, 4]) == 3

def test_force_to_pandasable():
    """ Test the force_to_pandasable function """
    rv = np.random.uniform(size=100)
    assert arrayUtils.forceToPandables(4) == 4
    assert np.allclose(arrayUtils.forceToPandables(rv), rv)
    try:
        arrayUtils.forceToPandables(rv, 95)
    except ValueError:
        pass
    else:
        raise ValueError("Failed to catch array length mismatch in arrayUtils.forceToPandables")
    rv2d = rv.reshape(10, 10)
    rv2d_check = np.vstack(arrayUtils.forceToPandables(rv2d))
    assert np.allclose(rv2d, rv2d_check)

    rv3d = rv.reshape(10, 2, 5)
    rv3d_check = np.vstack(arrayUtils.forceToPandables(rv3d)).reshape(rv3d.shape)
    assert np.allclose(rv3d, rv3d_check)


def test_slice_dict():
    """ Test the slice_dict method """
    test_data = dict(scalar=np.random.uniform(size=10),
                     vector=np.random.uniform(size=100).reshape(10, 10),
                     mat=np.random.uniform(size=1000).reshape(10, 10, 10))
    sliced = arrayUtils.sliceDict(test_data, 1)
    assert np.allclose(sliced['scalar'], test_data['scalar'][1])
    assert np.allclose(sliced['vector'], test_data['vector'][1])
    assert np.allclose(sliced['mat'], test_data['mat'][1])

    mask = np.zeros((10), bool)
    mask[1] = True
    mask[4] = True
    sliced = arrayUtils.sliceDict(test_data, mask)
    assert np.allclose(sliced['scalar'], test_data['scalar'][mask])
    assert np.allclose(sliced['vector'], test_data['vector'][mask])
    assert np.allclose(sliced['mat'], test_data['mat'][mask])


def test_print_dict_shape():
    """ Test the print_dict_shape method """
    test_data = dict(scalar=np.random.uniform(size=10),
                     vector=np.random.uniform(size=100).reshape(10, 10),
                     mat=np.random.uniform(size=1000).reshape(10, 10, 10))
    arrayUtils.printDictShape(test_data)


def test_concatenateDicts():
    """ Test the print_dict_shape method """
    test_data = dict(scalar=np.random.uniform(size=10),
                     vector=np.random.uniform(size=100).reshape(10, 10),
                     mat=np.random.uniform(size=1000).reshape(10, 10, 10))
    od = arrayUtils.concatenateDicts([test_data, test_data])
    assert np.allclose(od['scalar'][0:10], test_data['scalar'])
    assert np.allclose(od['vector'][0:10], test_data['vector'])
    assert np.allclose(od['mat'][0:10], test_data['mat'])
    assert np.allclose(od['scalar'][10:], test_data['scalar'])
    assert np.allclose(od['vector'][10:], test_data['vector'])
    assert np.allclose(od['mat'][10:], test_data['mat'])


def test_getInit():

    test_data = dict(scalar=np.random.uniform(size=10),
                     vector=np.random.uniform(size=100).reshape(10, 10),
                     mat=np.random.uniform(size=1000).reshape(10, 10, 10))

    dd = arrayUtils.getInitializationForODict(test_data)
    assert dd['scalar'][0] == (10,)
    assert dd['vector'][0] == (10, 10)
    assert dd['mat'][0] == (10, 10, 10)
    dd = arrayUtils.getInitializationForODict(test_data, 12)
    assert dd['scalar'][0] == (12,)
    assert dd['vector'][0] == (12, 10)
    assert dd['mat'][0] == (12, 10, 10)
    

def test_types():
    """ Test the typing functions"""
    assert not types.istablelike(4)
    assert types.istablelike(dict(a=np.ones(4)))
    assert not types.istablelike(dict(a=np.ones(4), b=np.ones(5)))
    assert not types.istabledictlike(4)
    assert not types.istabledictlike(dict(a=np.ones(4)))
    assert types.istabledictlike(dict(data=dict(a=np.ones(4))))
    try:
        types.fileType('xx.out')
    except KeyError:
        pass
    else:
        raise KeyError("Failed to catch unknown fileType")



def test_lazy_load():
    """ Test that the lazy import works"""
    noModule = lazyImport('thisModuleDoesnotExist')
    try:
        noModule.d
    except ImportError:
        pass
    else:
        raise ImportError("lazyImport failed")



""" Lazy loading modules """

import sys
import importlib.util
import h5py

class DeferredModuleError:
    """ Class to throw an error if you try to use a modules that wasn't loaded """

    def __init__(self, moduleName):
        self._moduleName = moduleName

    @property
    def moduleName(self):
        """ Return the name of the module this is associated to """
        return self._moduleName

    def __getattr__(self, item):
        raise ImportError(f"Module {self.moduleName} was not loaded, so call to {self.moduleName}.{item} fails")



def lazyImport(modulename):
    """ This will allow us to lazy import various modules

    Parameters
    ----------
    modulename : `str`
        The name of the module in question

    Returns
    -------
    module : `importlib.LazyModule`
        A lazy loader for the module in question
    """
    try:
        return sys.modules[modulename]
    except KeyError:
        spec = importlib.util.find_spec(modulename)
        if spec is None:
            print(f"Can't find module {modulename}")
            return DeferredModuleError(modulename)
        module = importlib.util.module_from_spec(spec)
        loader = importlib.util.LazyLoader(spec.loader)
        # Make module with proper locking and get it inserted into sys.modules.
        loader.exec_module(module)
        try:
            _ = dir(module)
        except ValueError:
            pass
    return module


tables = lazyImport('tables')
import astropy.table as apTable # lazyImport('astropy.table')
fits = lazyImport('astropy.io.fits')
h5py = h5py  #pylint: disable=self-assigning-variable
pd = lazyImport('pandas')
pq = lazyImport('pyarrow.parquet')

HAS_TABLES = tables is not None
HAS_PQ = pq is not None
HAS_FITS = fits is not None
HAS_ASTROPY = apTable is not None
HAS_HDF5 = h5py is not None
HAS_PANDAS = pd is not None



"""
Utilities for testing
"""

import numpy as np
from astropy.table import Table as apTable
from astropy.utils.diff import report_diff_values


def compare_tables(t1, t2):
    """ Compare all the tables in two `astropy.table.Table`)

    Parameters
    ----------
    t1 : `astropy.table.Table`
       One table
    t2 : `astropy.table.Table`
       Another tables

    Returns
    -------
    identical : `bool`
        True if the tables are identical, False otherwise

    Notes
    -----
    For now this explicitly flattens each of the columns, to avoid issues with shape
    """
    if sorted(t1.colnames) != sorted(t2.colnames):  #pragma: no cover
        return False
    for cname in t1.colnames:
        c1 = t1[cname]
        c2 = t2[cname]
        if not np.allclose(np.array(c1).flat, np.array(c2).flat):  #pragma: no cover
            return False
    return True


def compare_table_dicts(d1, d2, strict=False):
    """ Compare all the tables in two `OrderedDict`, (`str`, `astropy.table.Table`)

    Parameters
    ----------
    d1 : `OrderedDict`, (`str`, `astropy.table.Table`)
       One dictionary of tables
    d2 : `OrderedDict`, (`str`, `astropy.table.Table`)
       Another dictionary of tables

    Returns
    -------
    identical : `bool`
        True if all the tables are identical, False otherwise
    """
    identical = True
    for k, v in d1.items():
        try:
            vv = d2[k]
        except KeyError:  #pragma: no cover
            vv = d2[k.upper()]
        if strict:  #pragma: no cover
            identical &= report_diff_values(v, vv)
        else:  #pragma: no cover
            identical &= compare_tables(v, vv)
    return identical


def make_test_data():
    """ Make and return some test data """
    nrow = 1000
    vect_size = 20
    mat_size = 5
    scalar = np.random.uniform(size=nrow)
    vect = np.random.uniform(size=nrow*vect_size).reshape(nrow, vect_size)
    matrix = np.random.uniform(size=nrow*mat_size*mat_size).reshape(nrow, mat_size, mat_size)
    data = dict(scalar=scalar, vect=vect, matrix=matrix)
    table = apTable(data)
    table.meta['a'] = 1
    table.meta['b'] = None
    table.meta['c'] = [3, 4, 5]
    small_table = apTable(dict(a=np.ones(21), b=np.zeros(21)))
    small_table.meta['small'] = True
    tables = dict(data=table, md=small_table)
    return tables



"""
Unit tests for io_layer module
"""

import os

import unittest

from tables_io import types, convert, TableDict
from tables_io.testUtils import compare_table_dicts, make_test_data


class TableDictTestCase(unittest.TestCase):  #pylint: disable=too-many-instance-attributes
    """ Test the utility functions """

    def setUp(self):
        """
        Make any objects that are used in multiple tests.
        """
        self._tables = make_test_data()
        self._files = []

    def tearDown(self):
        """ Clean up any mock data files created by the tests. """
        for ff in self._files:
            if os.path.exists(ff):
                os.unlink(ff)

    def _do_loopback(self, tType, basepath, fmt, keys=None):
        """ Utility function to do loopback tests """
        odict_c = convert(self._tables, tType)
        td_c = TableDict(odict_c)
        filepath = td_c.write(basepath, fmt)
        if keys is None:
            self._files.append(filepath)
        else:
            for key in keys:
                self._files.append("%s%s.%s" % (basepath, key, fmt))
        td_r = TableDict.read(filepath, tType=tType, fmt=fmt, keys=keys)
        tables_r = td_r.convert(types.AP_TABLE)
        assert compare_table_dicts(self._tables, tables_r)
        if fmt in ['pq', 'h5']:
            return
        basepath2 = "%s_v2" % basepath
        filepath2 = td_c.write(basepath2)
        self._files.append(filepath2)
        td_r2 = TableDict.read(filepath2, tType=tType)
        tables_r2 = td_r2.convert(types.AP_TABLE)
        assert compare_table_dicts(self._tables, tables_r2)

    def testFitsLoopback(self):
        """ Test writing / reading to FITS """
        self._do_loopback(types.AP_TABLE, 'test_td_out', 'fits')

    def testHf5Loopback(self):
        """ Test writing / reading astropy tables to HDF5 """
        self._do_loopback(types.AP_TABLE, 'test_td_out', 'hf5')

    def testHdf5Loopback(self):
        """ Test writing / reading numpy arrays to HDF5 """
        self._do_loopback(types.NUMPY_DICT, 'test_td_out', 'hdf5')

    def testH5Loopback(self):
        """ Test writing / reading pandas dataframes to HDF5 """
        self._do_loopback(types.PD_DATAFRAME, 'test_td_out', 'h5')

    def testPQLoopback(self):
        """ Test writing / reading pandas dataframes to parquet """
        self._do_loopback(types.PD_DATAFRAME, 'test_td_out', 'pq', list(self._tables.keys()))

    def testBadType(self):
        """ Test adding bad types to TableDict """
        td = TableDict(self._tables)
        try:
            td['aa'] = 4
        except TypeError:
            pass
        else:
            raise TypeError("Failed to catch TypeError")

if __name__ == '__main__':
    unittest.main()



"""Functions to store analysis results as astropy data tables  """

from collections import OrderedDict

from .ioUtils import tableType, write, read, convert


class TableDict(OrderedDict):
    """Object to collect various types of table-like objects

    This class is a dictionary mapping name to table-like
    and a few helper functions, e.g., to add new tables to the dictionary
    and to read and write files, either as FITS or HDF5 files.
    """


    def __setitem__(self, key, value):
        try:
            _ = tableType(value)
        except TypeError as msg:
            raise TypeError("item %s was not recognized as a table.") from msg

        return OrderedDict.__setitem__(self, key, value)

    def write(self, basepath, fmt=None):
        """ Write tables to the corresponding file type

        Parameters
        ----------
        basepath : `str`
            base path for output files.  Suffix will be added based on type
        fmt : `str` or `None`
            The output file format, If `None` this will use `writeNative`
        """
        return write(self, basepath, fmt)

    def convert(self, tType):
        """ Build a new TableDict by converting all the table in the object to a different type

        Parameters
        ----------
        tType : `int`
            The type to convert to

        Returns
        -------
        td : `TableDict`
            The new TableDict
        """
        return TableDict(convert(self, tType))


    @classmethod
    def read(cls, filepath, tType=None, fmt=None, keys=None):
        """ Read a file to the corresponding table type

        Parameters
        ----------
        filepath : `str`
            File to load
        tType : `int` or `None`
            Table type, if `None` this will use `readNative`
        fmt : `str` or `None`
            File format, if `None` it will be taken from the file extension
        keys : `list` or `None`
            Keys to read for parquet files

        Returns
        -------
        tableDict : `TableDict`
            The data
        """
        return cls(read(filepath, tType, fmt, keys))



"""
Unit tests for io_layer module
"""

import os

import unittest
from tables_io import types, convert, io_open, read, write, iterator
from tables_io.testUtils import compare_table_dicts, compare_tables, make_test_data
from tables_io.lazy_modules import apTable


class IoTestCase(unittest.TestCase):  #pylint: disable=too-many-instance-attributes
    """ Test the utility functions """

    def setUp(self):
        """
        Make any objects that are used in multiple tests.
        """
        self._tables = make_test_data()
        self._table = self._tables['data']
        self._files = []

    def tearDown(self):
        """ Clean up any mock data files created by the tests. """
        for ff in self._files:
            if os.path.exists(ff):
                os.unlink(ff)

    def _do_loopback(self, tType, basepath, fmt, keys=None):
        """ Utility function to do loopback tests """
        odict_c = convert(self._tables, tType)
        filepath = write(odict_c, basepath, fmt)
        if keys is None:
            self._files.append(filepath)
        else:
            for key in keys:
                self._files.append("%s%s.%s" % (basepath, key, fmt))
        odict_r = read(filepath, tType=tType, fmt=fmt, keys=keys)
        tables_r = convert(odict_r, types.AP_TABLE)
        assert compare_table_dicts(self._tables, tables_r)

    def _do_loopback_single(self, tType, basepath, fmt, keys=None):
        """ Utility function to do loopback tests """
        obj_c = convert(self._table, tType)
        filepath = write(obj_c, basepath, fmt)
        try:
            os.unlink(filepath)
        except FileNotFoundError:
            pass
        filepath_check = write(obj_c, filepath, None)
        assert filepath == filepath_check
        if keys is None:
            self._files.append(filepath)
        else:
            for key in keys:
                self._files.append("%s%s.%s" % (basepath, key, fmt))
        obj_r = read(filepath, tType=tType, fmt=fmt, keys=keys)
        table_r = convert(obj_r, types.AP_TABLE)
        assert compare_tables(self._table, table_r)

        if types.FILE_FORMAT_SUFFIXS[fmt] not in types.NATIVE_TABLE_TYPE:
            return

        basepath_native = "%s_native" % basepath
        filepath_native = write(obj_c, basepath_native)
        if keys is not None:
            filepath_native += ".pq"
        self._files.append(filepath_native)
        obj_r_native = read(filepath_native, tType=tType, keys=keys)
        table_r_native = convert(obj_r_native, types.AP_TABLE)
        assert compare_tables(self._table, table_r_native)


    def _do_iterator(self, filepath, tType, expectFail=False, **kwargs):
        """ Test the chunk iterator """
        dl = []
        try:
            for _, _, data in iterator(filepath, tType=tType, **kwargs):
                dl.append(convert(data, types.AP_TABLE))
            table_iterated = apTable.vstack(dl)
            assert compare_tables(self._table, table_iterated)
        except NotImplementedError as msg:
            if expectFail:
                pass
            else:
                raise msg

    def _do_open(self, filepath, in_context=True):
        """  """
        if not in_context:
            f = io_open(filepath)
            assert f
            return
        with io_open(filepath) as f:
            assert f
        
    def testFitsLoopback(self):
        """ Test writing / reading to FITS """
        self._do_loopback(types.AP_TABLE, 'test_out', 'fits')
        self._do_loopback_single(types.AP_TABLE, 'test_out_single', 'fits')
        self._do_iterator('test_out_single.fits', types.AP_TABLE, True)
        self._do_open('test_out_single.fits')
        self._do_open('test_out.fits')
        
    def testRecarrayLoopback(self):
        """ Test writing / reading to FITS """
        self._do_loopback(types.NUMPY_RECARRAY, 'test_out', 'fit')
        self._do_loopback_single(types.NUMPY_RECARRAY, 'test_out_single', 'fit')
        self._do_iterator('test_out_single.fit', types.NUMPY_RECARRAY, True)
        self._do_open('test_out_single.fit')
        self._do_open('test_out.fit')
        
    def testHf5Loopback(self):
        """ Test writing / reading astropy tables to HDF5 """
        self._do_loopback(types.AP_TABLE, 'test_out', 'hf5')
        self._do_loopback_single(types.AP_TABLE, 'test_out_single', 'hf5')
        self._do_iterator('test_out_single.hf5', types.AP_TABLE, True, chunk_size=50)
        self._do_open('test_out_single.hf5')
        self._do_open('test_out.hf5')
        
    def testHdf5Loopback(self):
        """ Test writing / reading numpy arrays to HDF5 """
        self._do_loopback(types.NUMPY_DICT, 'test_out', 'hdf5')
        self._do_loopback_single(types.NUMPY_DICT, 'test_out_single', 'hdf5')
        self._do_iterator('test_out_single.hdf5', types.NUMPY_DICT, False, chunk_size=50)
        self._do_open('test_out_single.hdf5')
        self._do_open('test_out.hdf5')
        
    def testH5Loopback(self):
        """ Test writing / reading pandas dataframes to HDF5 """
        self._do_loopback(types.PD_DATAFRAME, 'test_out', 'h5')
        self._do_loopback_single(types.PD_DATAFRAME, 'test_out_single', 'h5')
        self._do_iterator('test_out_single.h5', types.PD_DATAFRAME, True, chunk_size=50)
        self._do_open('test_out_single.h5')
        self._do_open('test_out.h5')
        
    def testPQLoopback(self):
        """ Test writing / reading pandas dataframes to parquet """
        self._do_loopback(types.PD_DATAFRAME, 'test_out', 'pq', list(self._tables.keys()))
        self._do_loopback_single(types.PD_DATAFRAME, 'test_out_single', 'pq', [''])
        self._do_iterator('test_out_single.pq', types.PD_DATAFRAME, True)
        self._do_open('test_out_single.pq', in_context=False)
        
    def testBad(self):  #pylint: disable=no-self-use
        """ Test that bad calls to write are dealt with """
        try:
            write('aa', 'null', 'fits')
        except TypeError:
            pass
        else:
            raise TypeError("Failed to catch unwritable type")
        assert write(False, 'null', 'fits') is None

if __name__ == '__main__':
    unittest.main()



from setuptools import setup

setup(
    name="tables_io",
    author="Eric Charles",
    author_email="charles@slac.stanford.edu",
    url = "https://github.com/LSSTDESC/tables_io",
    packages=["tables_io"],
    description="Input/output and conversion functions for tabular data",
    setup_requires=['setuptools_scm'],
    long_description=open("README.md").read(),
    package_data={"": ["README.md", "LICENSE"]},
    use_scm_version={"write_to":"tables_io/_version.py"},
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        ],
    install_requires=["numpy>=1.21.0;python_version>='3.8'",
                      "numpy<1.22.*;python_version<'3.8'",
                      "astropy;python_version>='3.8'",
                      "astropy<5.*;python_version<'3.8'",
                      "h5py",
                      "pandas;python_version>='3.8'",
                      "pandas<1.4.*;python_version<'3.8'",
                      "pyarrow",
                      "tables"]
)



"""tables_io is a library of functions for input, output and conversion of tabular data formats"""

try:
    from ._version import version
except:  #pylint: disable=bare-except   #pragma: no cover
    version = "unknown"  

from .lazy_modules import *

from . import types

from . import arrayUtils

from . import convUtils as conv

from . import ioUtils as io

from .tableDict import TableDict

convertObj = conv.convertObj

convert = conv.convert

writeNative = io.writeNative

write = io.write

readNative = io.readNative

read = io.read

io_open = io.io_open

iteratorNative = io.iteratorNative

iterator = io.iterator



"""
Unit tests for io_layer module
"""

import unittest
from tables_io import types, convert, convertObj
from tables_io.testUtils import compare_table_dicts, make_test_data


class ConvTestCase(unittest.TestCase):  #pylint: disable=too-many-instance-attributes
    """ Test the utility functions """

    def setUp(self):
        """
        Make any objects that are used in multiple tests.
        """
        self._tables = make_test_data()
        self._table = self._tables['data']

    def tearDown(self):
        """ Clean up any mock data files created by the tests. """

    def _do_loopback(self, tType1, tType2):
        """ Utility function to do loopback tests """
        odict_1 = convert(self._tables, tType1)
        odict_2 = convert(odict_1, tType2)
        tables_r = convert(odict_2, types.AP_TABLE)
        assert compare_table_dicts(self._tables, tables_r)
        t1 = convertObj(self._table, tType1)
        t2 = convertObj(t1, tType2)
        _ = convertObj(t2, types.AP_TABLE)

    def testAstropyLoopback(self):
        """ Test writing / reading astropy tables to HDF5 """
        self._do_loopback(types.AP_TABLE, types.NUMPY_DICT)
        self._do_loopback(types.AP_TABLE, types.NUMPY_RECARRAY)
        self._do_loopback(types.AP_TABLE, types.PD_DATAFRAME)

    def testNumpyLoopback(self):
        """ Test writing / reading numpy arrays to HDF5 """
        self._do_loopback(types.NUMPY_DICT, types.AP_TABLE)
        self._do_loopback(types.NUMPY_DICT, types.NUMPY_RECARRAY)
        self._do_loopback(types.NUMPY_DICT, types.PD_DATAFRAME)

    def testRecarrayLoopback(self):
        """ Test writing / reading numpy arrays to HDF5 """
        self._do_loopback(types.NUMPY_RECARRAY, types.NUMPY_DICT)
        self._do_loopback(types.NUMPY_RECARRAY, types.AP_TABLE)
        self._do_loopback(types.NUMPY_RECARRAY, types.PD_DATAFRAME)

    def testPandasLoopback(self):
        """ Test writing / reading pandas dataframes to HDF5 """
        self._do_loopback(types.PD_DATAFRAME, types.AP_TABLE)
        self._do_loopback(types.PD_DATAFRAME, types.NUMPY_DICT)
        self._do_loopback(types.PD_DATAFRAME, types.NUMPY_RECARRAY)


if __name__ == '__main__':
    unittest.main()
